'' =================================================================================================
''
''   File....... jm_activitybot_jbtc.spin
''   Purpose.... Drive Activity Bot with Joystick BT Commander
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2015 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started.... 
''   Updated.... 26 FEB 2015
''
'' =================================================================================================


con { timing }

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                           ' use 5MHz crystal

  CLK_FREQ = _xinfreq * (_clkmode >> 6)                          ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                    ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                                ' ticks in 1us
                                                                  
                                                                  
con { activity board io pins }                                    
                                                                  
  RX1     = 31                                                   ' programming / terminal
  TX1     = 30                                                    
                                                                  
  SDA     = 29                                                   ' eeprom / i2c
  SCL     = 28                                                    
                                                                  
  AUD_RT  = 27                                                   ' audio (d/a) outputs
  AUD_LF  = 26                                                    
                                                                  
  LED_R   = 27                                                   ' leds (on audio outputs)
  LED_L   = 26                                                    
                                                                  
  SD_CS   = 25                                                   ' usd card
  SD_DI   = 24                                                    
  SD_CLK  = 23                                                    
  SD_DO   = 22                                                    
                                                                  
  ADC_CS  = 21                                                   ' adc input
  ADC_CLK = 20                                                    
  ADC_DO  = 19 { miso }                                           
  ADC_DI  = 18 { mosi }                                           
                                                                  
  SVO_6   = 6                                                   ' servo headers
  SVO_5   = 7                                                   
                                                                  
  SVO_4   = 15                                                    
  SVO_3   = 14                                                    
                                                                  
  SVO_R   = 13                                                    
  SVO_L   = 12                                                    

  RX2     = 11                                                   ' to rn-42 tx pin (do for xb module)
  TX2     = 10                                                   ' to rn-42 rx pin (di for xb module)


con

  TRM_BAUD = 115_200
  BT_BAUD  = 115_200                                             ' must match bt uart configuration

  #0, NUL, SOH, STX, ETX, EOT, ENQ                               ' packet framing characters
  #0, SC, MN, HR                                                 ' soft rtc registers

  #0, NO, YES

  DEBUG = NO

  USE_ADC = NO


con

  S_STOP = 1520                                                  ' stop value for CR servo (us)
  S_MAX  =  100                                                  ' max speed delta from stop (us)

                               
obj

  term    : "jm_fullduplexserial"                                ' * serial for terminal
  btio    : "jm_fullduplexserial"                                ' * serial for rn-42
  adc     : "jm_adc124s021"                                      '   for onboard adc
  io      : "jm_io"                                              '   basic io
  time    : "jm_time"                                            '   delays/timing
  etimer  : "jm_time"                                            '   timer for emergency stop
  stimer  : "jm_time"                                            '   servo update timing
  dbtimer : "jm_time"                                             
                                                                  
' * uses cog when loaded
                                                      
                                                                  
var

  long  xpos                                                     ' signed x position value
  long  ypos                                                     ' signed y position value

  byte  buttons                                                  ' button bits
  byte  lastbuttons                                              ' for detecting chages

  byte  buf[20]                                                  ' raw input buffer from JBTC
  byte  bufidx                                                   ' buffer index
  byte  buflen                                                   ' # bytes in input buffer
  
  byte  data1[13]                                                ' data fields in JBTC display
  byte  data2[13]
  byte  data3[13] 

  byte  rtc[3]                                                   ' seconds, minuts, hours


dat

  TimeStr     byte    "00:00:00", 0                              


pub main | c, ms                                                
                                                                  
  setup                                                          ' start objects

  repeat

    ' process packet from bluetooth
  
    repeat
      c := btio.rxcheck                                          ' look for bt byte
      if (c < 0)                                                 ' exit if nothing there
        quit
      else
        if (c == STX)                                            ' start of packet?
          buf[0] := c
          bufidx := 1
        elseif (bufidx > 0)                                      ' in packet
          buf[bufidx++] := c                                     ' yes, add c to packet
          if (c == ETX)                                          ' end of packet marker?
            process_packet                                       ' yes, handle packet
            bufidx := 0                                          ' reset buffer index
            etimer.start                                         ' reset e-stop timer
            quit                                                 ' escape this loop

    if (etimer.millis => 1000)                                   ' no packets for 1 second?
      etimer.start                                               ' restart
      xpos := 0                                                  ' zero out joystick 
      ypos := 0
      buttons := %000000
      
        
    ' update servos 

    ms := stimer.millis                                          ' capture servo timer
    if (ms => 20)                                                ' at least 20ms elapsed?
      stimer.adjust(-20)                                         ' yes, back-up timer by 20
      update_servos                                              ' and refresh servos
      

    ' update rtc

    ms := time.millis                                            ' capture rtc timer
    if (ms => 1000)                                              ' at least 1000ms (1s) elapseed?
      time.adjust(-1000)                                         ' yes, back-up by 1s
      if (++rtc[SC] == 60)                                       ' incremetn seconds with rollover       
        rtc[SC] := 0                                                     
        if (++rtc[MN] == 60)                                     ' increment minutes with rollover
          rtc[MN] := 0            
          if (++rtc[HR] == 24)                                   ' increment hours with rollover                 
            rtc[HR] := 0          
          
      dec2(rtc[HR], @TimeStr[0])                                 ' convert rtc bytes to string
      dec2(rtc[MN], @TimeStr[3])
      dec2(rtc[SC], @TimeStr[6])
      bytemove(@data1, @TimeStr, 9)                              ' copy to data1 slot
      update_jbtc                                                ' send to jbtc

       
       
pub process_packet

  if (bufidx == 8)                                               ' joystick packet
    get_joystick 
    return

  if (bufidx == 3)                                               ' button packet
    get_button                                                   ' get new state(s)
    process_button                                               ' handle changes immediately
    return

  ' could add bad packet handler here
  

pub update_servos | lsvo, rsvo, tadj

'' Adjust servo speed from x, y values
'' -- uses global values xpos and ypos

  lsvo := ypos                                                   ' get throttle pos
  rsvo := ypos

  tadj := ||xpos                                                 ' xpos is turning adjustment
  if (tadj => 5)                                                 ' if not in deadband
    tadj := map(tadj, 5, 100, 0, 67)                             ' reduce turning sensitivity
    if (xpos > 0)                                                ' right turn?
      rsvo := (100 - tadj) * rsvo / 100                          ' slow right servo
    elseif (xpos < 0)                                            ' left turn
      lsvo := (100 - tadj) * lsvo / 100                          ' slow left servo

  lsvo := map(lsvo, -100, 100, S_STOP-S_MAX, S_STOP+S_MAX)       ' convert to servo pulse
  phsa := -lsvo * US_001                                         ' update servo
      
  rsvo := map(rsvo, -100, 100, S_STOP+S_MAX, S_STOP-S_MAX)   
  phsb := -rsvo * US_001


  if (DEBUG == YES)
    term.dec(lsvo)
    term.tx(term#TAB)
    term.dec(rsvo)
    term.tx(term#TAB)
    term.bin(buttons, 6)
    term.tx(term#CR)


pub map(value, inmin, inmax, outmin, outmax)

'' Maps value in range inmin..inmax to new value in range outmin..outmax
'' -- liberated from Arduino library
'' -- caution: no bounds checking

  return (value - inmin) * (outmax - outmin) / (inmax - inmin) + outmin


pub get_joystick

'' Process joystick packet from JBTC
'' -- JBTC sends joystick packet as <STX> <X100> <X10> <X1> <Y100> <Y10> <Y1> <ETX>
''    * x and y digits are ascii chars
  
  xpos := get_dec3(@buf[1])                                      ' extract x joystick
  ypos := get_dec3(@buf[4])                                      ' extract y joystick


pub get_dec3(p_buf) | value

'' Converts three ASCII chars to decimal value
'' -- JBTC sends values (as strings) from "100" to "300" to represent -100 to +100

  value := (byte[p_buf++] - "0") * 100                           ' ascii to decimal converstion
  value += (byte[p_buf++] - "0") * 10
  value += (byte[p_buf] - "0")
  value -= 200                                                   ' remove offset from value

  return value
  

pub get_button | b, mask

'' Process button packet from JBTC
'' -- JBTC sends values from "A" to "L"
''    * "A" = B1 on, "B" = B1 off, "C" = B2 on, etc.

  b := buf[1]
  if ((b < "A") or (b > "L"))                                    ' validate button event code
    return

  b -= "A"                                                       ' convert command char to decimal (0..11)    

  ' each button has two states (on and off) hence one mask for both
         
  mask := 1 << (b / 2)                                           ' create mask for button    
  if (b & 1)                                                     ' if bit0 set ("B", "D", "F", etc.)            
    buttons &= !mask                                             '  clear button bit           
  else                           
    buttons |= mask                                              ' else set button bit            


pub process_button | delta

'' Isolates button that changed to call handler
'' -- will pass 0 (off) or button mask (on) to handler

  delta := (buttons ^ lastbuttons) & %111111                     ' find change
  lastbuttons := buttons                                         ' save for next change

  case delta                                                     ' process changed input
    %000001 : b1_change(buttons & delta)                         ' pass new state as mask 
    %000010 : b2_change(buttons & delta)          
    %000100 : b3_change(buttons & delta)          
    %001000 : b4_change(buttons & delta)          
    %010000 : b5_change(buttons & delta)          
    %100000 : b6_change(buttons & delta)


pub b1_change(state)

  if (state)
    io.high(26)
  else
    io.low(26)

    
pub b2_change(state)

  if (state)
    io.high(27)
  else
    io.low(27)


pub b3_change(state)


pub b4_change(state)   
    

pub b5_change(state)


pub b6_change(state)                                             ' clear other buttons when B6 pressed

  if (state)
    buttons := %000000
    lastbuttons := %000000

    ' call events that could be effected by clearing buttons
    
    b1_change(0)
    b2_change(0)
    update_jbtc


pub update_jbtc

'' Sends button status and data fields (ascii) to jbtc

  btio.tx(STX)
  btio.bin(buttons, 6)                                           ' update buttons
  btio.tx($01)  
  btio.str(@data1)                               
  btio.tx($04)
  btio.str(@data2)                               
  btio.tx($05)
  btio.str(@data3)                               
  btio.tx(ETX)
   

pub dec2(val, p_str)

'' Convert val to 2-digit ascii string at p_str

  byte[p_str][0] := val  / 10 + "0"
  byte[p_str][1] := val // 10 + "0"

  return p_str                                                   ' for inline printing of str

  
pub setup

'' Setup objects for project

  if (DEBUG == YES)
    term.start(RX1, TX1, %0000, TRM_BAUD)                        ' start serial for terminal         

  btio.start(RX2, TX2, %0000, BT_BAUD)                           ' start serial for bluetooth

  if (USE_ADC == YES)                                                         
    adc.start(ADC_CS, ADC_CLK, ADC_DI, ADC_DO)                   ' connect to adc

  time.start                                                     ' rtc timer
  etimer.start                                                   ' e-stop timer
  stimer.start                                                   ' servo timer
  dbtimer.start                                                  ' debug timer

  ' configure counters for servo pulses

  frqa := 1                                                      ' set for hi-res pulse                
  phsa := -S_STOP * US_001                                       ' output stop pulse (1520us)                           
  ctra := (%00100 << 26) | SVO_L                                 ' activate counter
  io.low(SVO_L)                                                  ' make servo pin an output      
                      
  frqb := 1                    
  phsb := -S_STOP * US_001                             
  ctrb := (%00100 << 26) | SVO_R
  io.low(SVO_R)
  
  
  
dat { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}